program cumulative(input, output);
(* cumulative: cumulative probability distribution

  Tom Schneider
  network address: toms@alum.mit.edu


  1991
 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.08; (* of cumulative.p 1991 September 16
origin 1991 Sep 16 *)
(* end module version *)

(* begin module describe.cumulative *)
(*
name
   cumulative: cumulative probability distribution

synopsis
   cumulative(input: in, output: out)

files
   input:  input file
   output: input file with line numbers and cumulative probability
     at the start

description

   Add line numbers to the input file along with cumulative probability.

   The input file should be sorted numerically first, for example with:

     cat myfile | sort -n -r | cumulative

examples

documentation

see also

   number.p

author
   Thomas Dana Schneider

bugs

technical notes

   The technical problem of number.p was solved here by first copying
   to an internal file.

*)
(* end module describe.cumulative *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline version = 4.11; (@ of prgmod.p 1991 Apr 22 *)

(* begin module cumulative.themain *)
procedure themain(var fin,fout: text);
(* the main procedure of the program *)
const
   decimals = 5; (* that should be sufficient for most files! *)
   width = 10;   (* width of number for probabilities *)
var
   count: integer; (* count of the lines processed so far *)
   hold: text; (* internal holding file *)
   lines: integer; (* total count of the lines *)
{
   decimals: integer; (* decimal places to express the value of lines in *)
}
begin
{ this is crazy.  the extra reset prevents
this from being used in pipes.  
   (* determine number of lines in the file *)
   lines := -1; (* don't count the line that has the eof on it! *)
   reset(fin);
   while not eof(fin) do begin
      lines := lines + 1;
      readln(fin)
   end;

   if lines > 0 then begin
      (* determine how many decimal places to put out for line numbers *)
      decimals := trunc(ln(lines)/ln(10))+1;

      (* copy the file *)
      reset(fin);
      for count := 1 to lines do begin
         write(fout,count:decimals,' ');
         copyaline(fin,fout)
      end;
   end;
   reset(fin);
}

   (* the strategy is that it would be nice to use this program
      in a pipe, which means that it cannot reset the input file.
      The (possible) solution is to copy the file to an internal
      file while counting the lines and then to spit that to the output. *)
    
   (* First copy to the internal file while counting *)
   lines := 0;
   rewrite(hold);
   while not eof(fin) do begin
      lines := succ(lines);
      copyaline(fin,hold)
   end;

   reset (hold); (* that's allowed because it is not in the pipe! *)

   count := 0;
   while not eof(hold) do begin
      count := succ(count);
      write(fout,     count:width,
                 ' ',(count/lines):width:decimals, (* ie the probability *)
                 ' ');
      copyaline(hold,fout)
   end;

end;
(* end module cumulative.themain *)

begin
   themain(input,output);
1: end.
