program fico(input, ficop, output);
(* fico: find control characters in the input file

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.01; (* of fico.p 2012 Oct 18
2012 Oct 18, 1.01: improve documentation, give more data to output
2012 Sep 25, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.fico *)
(*
name
   fico: find control characters in the input file

synopsis
   fico(input: in, ficop: in, output: out)

files

   input:  input text file

   ficop:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      Not used yet.  Could be used to set the cutoff.

   output: messages to the user

description

   Identify control characters in a text file.

examples

documentation

see also

author

   Thomas Dana Schneider

bugs

technical notes

   The constant HIGH: high order bit is on if greater than this value

*)
(* end module describe.fico *)

var
   ficop: (* file used by this program *)
    text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module fico.themain *)
procedure themain(var fin, ficop: text);
(* the main procedure of the program *)
const
   HIGH = 127; (* high order bit is on if greater than this value *)
var
   c: char; (* a character *)
   controlsfound: integer; (* number of control characters found *)
   linenumber: integer; (* of all the characters *)
   parameterversion: real; (* parameter version number *)
   s: char; (* an s (or not) to put at the end of a word *)
begin
   writeln(output,'fico ',version:4:2);
   writeln(output, 'Characters with ord > ', HIGH:1,' will be shown.');
{
   reset(ficop);
   readln(ficop, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
}

   linenumber := 1;
   controlsfound := 0;
   while not eof(fin) do begin
      if eoln(fin)
      then begin
         linenumber := succ(linenumber);
         readln(fin);
      end
      else begin
         read(fin,c);
         if ord(c) > HIGH then begin
            writeln(output,'line ', linenumber:1);
            controlsfound := controlsfound + 1;
         end;
      end;
   end;
   writeln(output,'number of lines in the input file: ', linenumber:1);
   if controlsfound = 1
   then s := ' '
   else s := 's';
   writeln(output,controlsfound:1,' control character',s,' found in input file');

end;
(* end module fico.themain *)

begin
   themain(input, ficop);
1: end.
