program logscale(input, output);
(* logscale: provide numbers on a log scale

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.02; (* of logscale.p 2010 Sep 18
2010 Sep 18, 1.02: remove unnecessary code
2010 Sep 18, 1.01: add width, decimals control
2010 Sep 18, 1.00: origin, functional
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.logscale *)
(*
name
   logscale: provide numbers on a log scale

synopsis
   logscale(input: in, output: out)

files

   input: one line with the following values:

      x (real): starting value
      b (integer): base of the scale (eg 2 or 10)
      n (integer): number of intervals
        If n is negative, report the program version number.
      m (integer): which interval we are on
      w (integer): width of output number (characters)
      d (integer): decimal places of output number (characters)

   output: messages to the user

description

   Create numbers on a log scale one at a time for use in scripts.

Let the scale look like this (where b = 10):

   x * z * z * z * z = 10x

   So here x is the starting value.  It is successively multiplied by z
   n = 4 times until we reach 10x.

   So for

   m = 0  result = x
   m = 1  result = x * z
   m = 2  result = x * z * z
   m = 3  result = x * z * z * z
   m = 4  result = x * z * z * z * z

   So now all we need to know is the value of z.

   From the first equation,

   z^n = b

   so

   z = b ^ (1/n)

   then:

   result = x * z^m

examples

The tcsh script:

set m = 0
while ($m < 13)
   # x (real): starting value
   # b (integer): base of the scale (eg 2 or 10)
   # n (integer): number of intervals
   # m (integer): which interval we are on
   # w (integer): width of output number (characters)
   # d (integer): decimal places of output number (characters)
   printf "%2d " $m
   #     x b n  m w d
   echo "1 2 4 $m 4 2" | logscale
   @ m = $m + 1
end

gives:

 0 1.00
 1 1.19
 2 1.41
 3 1.68
 4 2.00
 5 2.38
 6 2.83
 7 3.36
 8 4.00
 9 4.76
10 5.66
11 6.73
12 8.00

   output: messages to the user

description

   Create numbers on a log scale one at a time for use in scripts.

Let the scale look like this (where b = 10):

   x * z * z * z * z = 10x

   So here x is the starting value.  It is successively multiplied by z
   n = 4 times until we reach 10x.

   So for

   m = 0  result = x
   m = 1  result = x * z
   m = 2  result = x * z * z
   m = 3  result = x * z * z * z
   m = 4  result = x * z * z * z * z

   So now all we need to know is the value of z.

   From the first equation,

   z^n = b

   so

   z = b ^ (1/n)

   then:

   result = x * z^m

examples
   echo "

documentation


documentation

see also

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.logscale *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module logscale.themain *)
procedure themain;
(* the main procedure of the program *)
var
   x: real; (* starting value *)
   b: integer; (* base of the scale (eg 2 or 10) *)
   n: integer; (* number of intervals *)
   m: integer; (* which interval we are on *)
   z: real; (* interval on log scale *)
   w: integer; (* width of output numbers (characters) *)
   d: integer; (* decimal places of output numbers (characters) *)
   result: real; (* value to output *)
begin

   readln(input, x, b, n, m, w, d);

   if b < 1 then begin
      writeln(output,'base of scale must be greater than 1');
   end;

   if w < 0 then begin
      writeln(output,'width must be positive');
   end;

   if d < 0 then begin
      writeln(output,'decimals must be positive');
   end;

   (* If n is negative, report the program version number. *)
   if n < 0 then begin
      writeln(output,'logscale ',version:4:2);
   end
   else begin
      (* z = b ^ (1/n)
         compute this as:
      exp(ln(z)) = exp(ln(b)/n); *)
      z := exp(ln(b)/n);

      (* result := x * z^m *)
      result := x * exp(m*ln(z));

      writeln(output, result:w:d);
   end;

end;
(* end module logscale.themain *)

begin
   themain;
1: end.
