program mkdis(distribution, mkdisp, data, output);
(* mkdis: define a distribution

  Tom Schneider
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.02; (* of mkdis.p 1997 March 21
origin 1997 March 21 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.mkdis *)
(*
name
   mkdis: define a distribution

synopsis
   mkdis(distribution: in, mkdisp: in, data: out, output: out)

files
   distribution:  The first line of the file is the x value (integer)
      All remaining lines are y values.  Lines that begin with an
      asterisk '*' are skipped as comments.

   data:  the distribution is expanded.

   mkdisp:  parameters to control the program.  The file must contain the
        following parameters, one per line:

      0. The version number of the program.  This allows the user to be
         warned if an old parameter file is used.

   output: messages to the user

description
   If one is given a distribution, it would be nice to run it through
   the genhis program to get the average.  This program takes
   a distribution of integers and produces the raw data that would
   have generated that distribution.  The raw data can be run through
   genhis.  Genhis will reconstitute the distribution and give
   statistics about it.

examples
   For the distribution input:

   5
   3
   2
   7

   meaning that there are 3 cases of 5, 2 cases of 6 and 7 cases of 7.

   The distribution file is

         5
         5
         5
         6
         6
         7
         7
         7
         7
         7
         7
         7

   The distibution output by genhis is:

    5.00000          3   +++
    6.00000          2   ++
    7.00000          7   +++++++

documentation

see also
   genhis.p

author
   Thomas Dana Schneider

bugs

technical notes
   This is a silly but simple way to solve the problem.
   It is silly because the data are expanded.
   It is simple because this program was a snap to write!

*)
(* end module describe.mkdis *)

var
   distribution, (* file used by this program *)
   mkdisp, (* file used by this program *)
   data: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module mkdis.themain *)
procedure themain(var distribution, mkdisp, data: text);
(* the main procedure of the program *)
var
   parameterversion: real; (* parameter version number *)
   d: integer; (* a data item from data *)
   dindex: integer; (* index to d *)
   position: integer; (* position of d in the distribution *)
begin
   writeln(output,'mkdis ',version:4:2);
   reset(mkdisp);
   readln(mkdisp, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   reset(distribution);
   rewrite(data);

   (* skip any header *)
   while (distribution^ = '*') or eoln(distribution)
   do readln(distribution);
   readln(distribution, position);

   while not eof(distribution) do begin
      if (distribution^ = '*') or eoln(distribution)
      then readln(distribution)
      else begin
         readln(distribution, d);
         for dindex := 1 to d
         do writeln(data, position);
         position := succ(position)
      end;
   end;

end;
(* end module mkdis.themain *)

begin
   themain(distribution, mkdisp, data);
1: end.
