program notex(input,output);
(* notex: remove tex and latex constructs
   by Thomas Schneider, copyright (c) 1991  
module libraries required: delman, prgmods *)
  
label 1; (* end of program *) 
  
const 
(* begin module version *)
version = 1.32; (* of notex.p 1991 February 1
origin 1991 Feb 1 from untex *)
(* end module version *)
  
(* begin module describe.notex *) 
(*
name
   notex: remove tex and latex constructs
  
synopsis
   notex(input: in, output: out) 
  
files 
   input: a tex or latex file
   output: the file with:
      '\xxx' command words converted to spaces,
      '{$}' converted to spaces
      free floating '.' ',' '(' ')' removed
      comments (%) removed

      multiple spaces are comressed to single spaces.
      multiple lines are compressed to 2 lines (to preserve the
      paragraph structure).

      Only characters numbers and blanks are left behind

description 
   This reduces the number of words counted by wc to something close to correct.
   It is harsher than untex in that it specifically filters out
   everything except numbers, alphabetic chracters and the blank.

author
   Thomas D. Schneider 
  
bugs
   citations and comments on lines by themselves leave a blank line.
  
*)
(* end module describe.notex *)

(* begin module themain *)
procedure themain(var fin, fout: text);  
(* the main procedure of the program *)
var 
   blanks: integer; (* number of blanks stored up so far *)
   ch: char; (* a character read from fin *)  
   done: boolean; (* done killing a command *)
   lines: integer; (* count of number of blank lines stored up so far *)
procedure out(c: char);
(* output the character c.  reduce the blanks before c.
'\' means a carriage return is to be output *)
begin

   if c = ' '
   then blanks := blanks + 1
   else if c <> '\' then begin
      (* output any lines that were stored *)
      if lines > 0 then begin
         if lines > 1 then writeln(fout); (* paragraph representation *)
         writeln(fout);
         lines := 0; (* reset stored lines to zero if characters written *)
         blanks := 0; (* wipe out the previous blanks *)
      end;

      if blanks > 0 then write(fout,' ');
      write(fout,c);
      blanks := 0;
   end
   else begin (* store up blank lines *)
      lines := lines + 1;
   end
end;
begin 
   blanks := 0;
   lines := 0;
   while not eof(fin) do begin 
      while not eoln(fin) do begin
         read(fin,ch);

         (* bracketed text and equations *)
         if ch in ['{','$','}']
         then out(' ')

         (* free floating commas or periods eg after "\cite{...}." and parens *)
         else if (ch in ['.',',','(',')']) and (blanks>0)
         then begin (* nothing! *) end

         (* commands *)
         else if ch = '\'
         then begin (* delete the whole command *)
            out(' '); (* replace it with a space *)
            read(fin,ch); (* look at the next character *)
            while (not eoln(fin)) and
                  (ch in
                  ['a','b','c','d','e','f','g','h','i','j','k','l',
                   'm','n','o','p','q','r','s','t','u','v','w','x','y','z',
                   'A','B','C','D','E','F','G','H','I','J','K','L',
                   'M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z'])
            do read(fin,ch);
            if ch in ['{','['] then begin (* oh oh an argument to delete! *)
               repeat
                  read(fin,ch);
                  done :=  (ch = '}') or eof(fin);
                  if not done
                  then if eoln(fin) then readln(fin) (* to next line *)
               until done
            end
         end

         (* comments *)
         else if ch = '%' then begin(* delete comment line *)
            (* note: \% is removed by the previous step *)
            while not eoln(fin) do read(fin,ch)
         end

         (* other text and final cleaning *)
         else if ch in
                  ['a','b','c','d','e','f','g','h','i','j','k','l',
                   'm','n','o','p','q','r','s','t','u','v','w','x','y','z',
                   'A','B','C','D','E','F','G','H','I','J','K','L',
                   'M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
                   '0','1','2','3','4','5','6','7','8','9',
                   ' ']
              then out(ch)
              else out(' ');
      end;
      readln(fin);
      out('\');
   end
end;  
(* end module notex.themain *)
 
begin
   themain(input, output);
1: end.
