program rf(input,output);
(* calculate Rfrequency

  Dr. Thomas D. Schneider
  toms@alum.mit.edu
  permanent email: toms@alum.mit.edu
  https://alum.mit.edu/www/toms/

*)

const
(* begin module version *)
version = 1.08; (* of rf.p 2000 June 13
1.08, 2000 June 13: report frequency in exponential form
last change: 1988 October 14
origin 1986 feb 14 *)
(* end module version *)

(* begin module describe.rf *)
(*
name
   rf: calculate Rfrequency

synopsis
   rf(input: in, output: out)

files
   input: interactive input from the user
   output: messages to the user

description
   calculate Rfrequency as

   Rf = - log (number of binding sites / genome size)

   where the log is to the base 2, giving the result in bits.

documentation
   Schneider, T.D., G.D. Stormo, L. Gold and A. Ehrenfeucht (1986)
   The information content of binding sites on nucleotide sequences.
   J. Mol. Biol. 188: 415-431.

author
   thomas d. schneider

bugs
   The program depends on interactive procedures on a particular computer
   and so may need to be modified on transportation.

*)
(* end module describe.rf *)

var gamma,g: real; (* number of sites and genome size *)
    f: real; (* frequency of sites *)
begin
   writeln(output,'rf ',version:4:2);

   write(output,'number of potential binding sites, G: ');
   readln(input,g);

   write(output,'number of sites in the genome, gamma: ');
   readln(input,gamma);

   f:=gamma/g;
   writeln(output,'f = ',f:10);
   writeln(output,'Rfrequency = ',-ln(f)/ln(2):5:2,' bits');
end.
