program roundprogram(fin, roundp, fout, output);
(* round: round a set of numbers

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

*)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.01; (* of round.p 2016 Jan 25
2016 Jan 25, 1.01: avoid name conflict of program with function round.
2002 Feb  6, 1.00: origin
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.round *)
(*
name
   round: round a set of numbers

synopsis
   round(fin: in, roundp: in, fout: out, output: out)

files

   fin:  a set of real numbers, one per line, possibly followed by text.

   fout:  the real numbers of fin rounded, followed by the same text.

   roundp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      The second parameter is the number of decimal places to round to.

description

   Round numbers in a list.

examples

documentation

see also

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.round *)

var
   fin, (* file used by this program *)
   roundp, (* file used by this program *)
   fout: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module copyaline *)
procedure copyaline(var fin, fout: text);
(* copy a line from file fin to file fout *)
begin (* copyaline *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
      readln(fin);
      writeln(fout);
end; (* copyaline *)
(* end module copyaline from prgmod *)

(* begin module round.themain *)
procedure themain(var fin, roundp, fout: text);
(* the main procedure of the program *)
var
   decimals: integer; (* decimal places of output number *)
   number: real; (* a number to process *)
   parameterversion: real; (* parameter version number *)
begin
   writeln(output,'round ',version:4:2);
   reset(roundp);
   readln(roundp, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   readln(roundp, decimals);

   reset(fin);
   rewrite(fout);
   while not eof(fin) do begin
      read(fin, number);
      write(fout, number:(decimals+2):decimals,' ');
      copyaline(fin,fout);
   end;

end;
(* end module round.themain *)

begin
   themain(fin, roundp, fout);
1: end.
