program short(fin, fout, shortp, output);
(* short: find locations of short lines in a file
   tom schneider
   copyright (c) 1985
   module libraries required: delman, prgmods *)

label 1; (* the end of the program *)

const
(* begin module version *)
version = 1.01; (* of short 1985 may 30
origin: 1985 may 30 *)
(* end module version *)

(* begin module describe.short *)
(*
name
      short: find locations of short lines in a file

synopsis
      short(fin: in, fout: out, shortp: in, output: out)

files
      fin: the file to be analyzed
      fout: a list of lines that are short.
      shortp: a parameter to determine what 'short' means.
         this is one integer.  lines of this length or shorter
         will be reported to fout.
      output: messages to the user

description
      database programs that scan a line and assume that there are
      a certain number of characters on the line will lose track of
      the correct location if the line is shorter than they expect.
      this has happened with delila and dbcat.  the short program
      scans a file for lines shorter than a given length and
      lists them in the fout file.  the purpose of the program is
      to help debug database programs.


author
      thomas schneider

bugs
      none known
*)
(* end module describe.short *)

var
(* begin module short.var *)
      fin, (* input file *)
      fout, (* list of short lines *)
      shortp: (* parameters *)
         text;
(* end module short.var *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'prgmod 3.97  85 may 5 tds'; *)

(* begin module short.themain *)
procedure themain(var fin, fout, shortp: text);
(* the main procedure of short *)
var
      count: integer; (* number of characters on a line in fin *)
      detection: integer; (* the maximum length of a line to be detected *)
      line: integer; (* the number of a line in fin *)
begin
      writeln(output,' short ',version:4:2);
      reset(fin);
      rewrite(fout);
      writeln(fout,' short ',version:4:2);
      reset(shortp);

      if eof(shortp) then begin
         writeln(output,' shortp is empty');
         halt
      end
      else readln(shortp, detection);

      line := 0;
      while not eof(fin) do begin
         count := 0;
         while not eoln(fin) do begin
            get(fin);
            count := count + 1;
         end;
         readln(fin);
         line := line + 1;
         if count <= detection
               then writeln(fout,' at ',line:1,' count is ',count:1);
      end
end;
(* end module short.themain *)

begin
      themain(fin,fout,shortp);
1: end.
