program timefpc(output);
(* timefpc: test time on the FPC compiler

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms 

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.03; (* of timefpc.p 2016 Jul 26
2016 Jul 26, 1.03: remove weird 'output: datetimearray;' definition!
2016 Jul 14, 1.02: absorb more code: extpas.pp (failed)
2016 Jul 14, 1.01: absorb more code: withDate.inc which was inserted using
                   {$ifdef FPC}
                     {$i withDate.inc}
                   {$endif}
2016 Jul 14, 1.00: origin from timegpc.p
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.timefpc *)
(*
name
   timefpc: test time on the Free Pascal (fpc) compiler

synopsis
   timefpc(output: out)

files

   output: show the current time.

description

   The time function returned by various computers is different.  Show the
   time function to figure out how to modify the getdatetime procedure.

   This program contains routines for obtaining the date and time
   for Delila programs compiled with the fpc compiler.

   They can be easily transferred to other programs using the module
   program.  The script upgrademodule will do this.

   NOTE:  For now the file extpas.pp must be compiled by fpc this way:

       fpc -ap extpas.pp

   The resulting extpas.o and extpas.ppu must be in the same
   directory as the program to be compiled.

   Programs can then be compiled with
       fpc -Miso -ap -Faextpas PROGRAM.p

   The fpcc script does all these steps.

examples

   Times should look like this:
       1980/06/09 18:49:11
         ye mo da ho mi se

documentation

see also

   {The module program, used to transfer the modules is:} module.p

   {The FPC compiler is described at:}

   http://www.freepascal.org/

   {The FPC time routines are described in}

   http://www.freepascal.org/docs-html/rtl/sysutils/datetimeroutines.html

   {see the section: TIME HANDLING ROUTINES}

   {To compile Delila programs on the FPC compiler, use the} fpcc {script.}

   { PROBABLY THIS IS IRRELEVANT:
    Until fpcc gets smarter, for this program it is necessary to use the -time
   flag:
      fpcc timefpc.p -time}

   {Related date/time programs:}
   timesun.p
   timep2c.p
   timegpc.p

   {General discussion on compiling Delila programs:}
   https://alum.mit.edu/www/toms/delila.html#How.To.Compile

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.timefpc *)

(* begin module datetime.const *)

   datetimearraylength = 19; (* length of dataarray for dates,
       It is just long enough to include the 4 digit year - solving the
       year 2000 problem:
1980/06/09 18:49:11
123456789 123456789 
         1         2
*)

(* end module datetime.const version = 7.40; {of delmod.p 2000 Feb 18} *)

type

(* begin module datetime.type *)
   (* array for dates *)
   datetimearray = packed array[1..datetimearraylength] of char;
(* end module datetime.type version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module package.datetime *)
(* ************************************************************************ *)
(* begin module getdatetime *)
(* end module getdatetime *)

(* begin module readdatetime *)
(* end module readdatetime *)

(* begin module writedatetime *)
(* end module writedatetime *)

(* begin module timeseed *)    
(* end module timeseed *)
 
(*[[*)
(* begin module limitdate *)
(* end module limitdate *)
(*]]*)

(* ************************************************************************ *)
(* end module package.datetime *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module extpas.pp *)
{This would replace the external extpas.pp ... but it fails
    This file is part of the Free Pascal Run time library.
    Copyright (c) 2015 by Florian Klaempfl

    This unit contain procedures specific for an extended pascal
    compatible mode.  It should be platform independent.

    See the file COPYING.FPC, included in this distribution,
    For details about the copyright.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

 **********************************************************************

    type
      TimeStamp = packed record
        DateValid : Boolean;
        TimeValid : Boolean;
        year : Integer;
        month : 1..12;
        day : 1..31;
        hour : 0..23;
        minute : 0..59;
        second : 0..59;
        (* implementation specific *)
        second100 : 0..99;
      end;

    procedure GetTimeStamp(var ts : TimeStamp);
      var
        year1,month1,mday1,wday1,
        year2,month2,mday2,wday2,
        hour,minute,second,sec100 : word;
      begin
        repeat
          GetDate(year1,month1,mday1,wday1);
          GetTime(hour,minute,second,sec100);
          GetDate(year2,month2,mday2,wday2);
        (* the date may not have changed while we read the time *)
        until (year1=year2) and (month1=month2) and (mday1=mday2);
        ts.DateValid:=true;
        ts.TimeValid:=true;
        ts.year:=year1;
        ts.month:=month1;
        ts.day:=mday1;
        ts.hour:=hour;
        ts.minute:=minute;
        ts.second:=second;
        ts.second100:=sec100;
      end;
}
(* end module extpas.pp *)

(* begin module getdatetime *)

(*******************************************************************************)
(*******************************************************************************)
(* This code was originally 
withDate.inc
inserted using
> {$ifdef FPC}
>   {$i withDate.inc}
> {$endif}
*)

(* This redefinition is unnecessary:
type
  DateString = datetimearray;
*)

procedure date(var adatetime: datetimearray; Format: datetimearray); overload;
var
  i: 1..datetimearraylength + 1;
  j: 1..maxint;
  MyTS: TimeStamp;
  YearString:   packed array[1..4] of char;
  MonthString:  packed array[1..2] of char;
  DayString:    packed array[1..2] of char;
  HourString:   packed array[1..2] of char;
  MinuteString: packed array[1..2] of char;
  SecondString: packed array[1..2] of char;
begin
  GetTimeStamp(MyTS);
  YearString[1]   := char(ord('0') +  MyTS.year   div 1000);
  YearString[2]   := char(ord('0') + (MyTS.year   div 100) mod 10);
  YearString[3]   := char(ord('0') + (MyTS.year   div 10 ) mod 10);
  YearString[4]   := char(ord('0') +  MyTS.year   mod 10);
  MonthString[1]  := char(ord('0') +  MyTS.month  div 10);
  MonthString[2]  := char(ord('0') +  MyTS.month  mod 10);
  DayString[1]    := char(ord('0') +  MyTS.day    div 10);
  DayString[2]    := char(ord('0') +  MyTS.day    mod 10);
  HourString[1]   := char(ord('0') +  MyTS.hour   div 10);
  HourString[2]   := char(ord('0') +  MyTS.hour   mod 10);
  MinuteString[1] := char(ord('0') +  MyTS.minute div 10);
  MinuteString[2] := char(ord('0') +  MyTS.minute mod 10);
  SecondString[1] := char(ord('0') +  MyTS.second div 10);
  SecondString[2] := char(ord('0') +  MyTS.second mod 10);
  i := 1;
  j := 1;
  repeat
    if Format[i] <> '%' then
      adatetime[j] := Format[i]
    else
    begin
      system.inc(i);
      case Format[i] of
        'Y': begin
               adatetime[j]   := YearString[1];
               adatetime[j+1] := YearString[2];
               adatetime[j+2] := YearString[3];
               adatetime[j+3] := YearString[4];
               system.inc(j, 4);
             end;
        'm': begin
               adatetime[j]   := MonthString[1];
               adatetime[j+1] := MonthString[2];
               system.inc(j, 2);
             end;
        'd': begin
               adatetime[j]   := DayString[1];
               adatetime[j+1] := DayString[2];
               system.inc(j, 2);
             end;
        'H': begin
               adatetime[j]   := HourString[1];
               adatetime[j+1] := HourString[2];
               system.inc(j, 2);
             end;
        'M': begin
               adatetime[j]   := MinuteString[1];
               adatetime[j+1] := MinuteString[2];
               system.inc(j, 2);
             end;
        'S': begin
               adatetime[j]   := SecondString[1];
               adatetime[j+1] := SecondString[2];
               system.inc(j, 2);
             end;
      end;
    end;
    system.inc(i);
    system.inc(j);
  until i > datetimearraylength;
end;

procedure date(var adatetime: datetimearray); overload;
begin
  date(adatetime, '%Y/%m/%d');
end;

procedure time(var adatetime: datetimearray);
begin
  date(adatetime, '%H:%M:%S');
end;
(*******************************************************************************)
(*******************************************************************************)

procedure getdatetime(var adatetime: datetimearray);
(* get the date and time into a single array from the system clock.
   adatetime contains the date:
       1980/06/09 18:49:11
         ye mo da ho mi se
(year, month, day, hour, minute, second).
As of 2000 February 18, the Sun Pascal compiler requires a formatting
statement.  This statement allows the date to be generated in this
standard Delila format in a single call.  Information about the
formatting statement is available on the manual page for date in Unix.
If a computer does not have this method, see the 'oldgetdatetime' routine
in delmod.p (https://alum.mit.edu/www/toms/delila/delmod.html)
for some conversion code.

FPC Functions:
function  GetUnixTime (var MicroSecond : Integer) : UnixTimeType;

http://agnes.dida.physik.uni-essen.de/~gnu-pascal/gpc_109.html#SEC109

7.10.8 Date And Time Routines 

procedure GetTimeStamp (var t : TimeStamp); 
function Date (t : TimeStamp) : packed array [1 .. DateLength] of Char; 
function Time (t : TimeStamp) : packed array [1 .. TimeLength] of Char; 

DateLength and TimeLength are implementation dependent constants. 

GetTimeStamp (t) fills the record `t' with values. If they are valid, the Boolean
flags are set to True. 

TimeStamp is a predefined type in the Extended Pascal standard. It may be
extended in an implementation, and is indeed extended in FPC. For the full
definition of `TimeStamp', see section 8.255 TimeStamp. 

*)
var
    t: TimeStamp; 

(* begin module pluckdigit *)
function pluckdigit(number, logplace:integer): char;
(* return the digit at the place value ('logplace') position of number.
example:
      pluckdigit(13625, 3) = 3
      pluckdigit(13625, 4) = 1
This routine was taken from module numberdigit in prgmod.p, but is
modified so as not to give the sign.  Instead it gives zeros
above the digits.  'myabsolute' replaced 'absolute', which
is apparently a keyword for FPC.  The name is kept for
to keep the code looking similar to its origin.
*)
var
      place: integer; (* the exponent of logplace *)
      count: integer; (* used to make place *)
      myabsolute: integer; (* the absolute value of number *)
      acharacter: char; (* the character to be returned *)
procedure digit;
(* extract a digit at the place position *)
var
      tenplace: integer; (* ten times place *)
      z: integer; (* an intermediate value *)
      d: integer; (* the digit extracted *)
begin (* digit *)
      tenplace:=10*place;
      z:=myabsolute-((myabsolute div tenplace)*tenplace);
      if place = 1
         then d:=z
         else d:= z div place;
      case d of
         0: acharacter:='0';
         1: acharacter:='1';
         2: acharacter:='2';
         3: acharacter:='3';
         4: acharacter:='4';
         5: acharacter:='5';
         6: acharacter:='6';
         7: acharacter:='7';
         8: acharacter:='8';
         9: acharacter:='9';
      end
end; (* digit *)
begin (* pluckdigit *)
      place:=1;
      for count:=1 to logplace do place:=10*place;
      if number=0 then begin
         acharacter:='0'
      end
      else begin
         myabsolute:=number;
         if myabsolute >= place
         then digit
         else acharacter := '0'
      end;
      pluckdigit:=acharacter
end; (* pluckdigit *)
(* end module pluckdigit *)

begin

 (* according to:
http://agnes.dida.physik.uni-essen.de/~gnu-pascal/gpc_109.html#SEC109
*)
   GetTimeStamp(t); 

(* Predefined time stamp:
http://agnes.dida.physik.uni-essen.de/~gnu-pascal/gpc_389.html#SEC389
  TimeStamp = {@@packed} record
      DateValid,
      TimeValid   : Boolean;
      Year        : Integer;
      Month       : 1 .. 12;
      Day         : 1 .. 31;
      DayOfWeek   : 0 .. 6;  { 0 means Sunday }
      Hour        : 0 .. 23;
      Minute      : 0 .. 59;
      Second      : 0 .. 61; { to allow for leap seconds }
      MicroSecond : 0 .. 999999
    end;
*)

  with t do begin
    if TimeValid then begin
{
       writeln(output,'valid time');
       writeln(output,'year   =',year:4);
       writeln(output,'month  =',month:2);
       writeln(output,'day    =',day:2);
       writeln(output,'hour   =',hour:2);
       writeln(output,'minute =',minute:2);
       writeln(output,'second =',second:2);
}
       adatetime := 'year/mm/dd hh:mm:ss';
       adatetime[ 1] := pluckdigit(Year,3);
       adatetime[ 2] := pluckdigit(Year,2);
       adatetime[ 3] := pluckdigit(Year,1);
       adatetime[ 4] := pluckdigit(Year,0);
       adatetime[ 6] := pluckdigit(Month,1);
       adatetime[ 7] := pluckdigit(Month,0);
       adatetime[ 9] := pluckdigit(Day,1);
       adatetime[10] := pluckdigit(Day,0);
       adatetime[12] := pluckdigit(Hour,1);
       adatetime[13] := pluckdigit(Hour,0);
       adatetime[15] := pluckdigit(Minute,1);
       adatetime[16] := pluckdigit(Minute,0);
       adatetime[18] := pluckdigit(Second,1);
       adatetime[19] := pluckdigit(Second,0);
    end
    else begin
       writeln(output,'getdatetime: invalid time!');
       halt;
    end
  end;
{
   Sun compiler method:
   date(adatetime,'%Y/%m/%d %H:%M:%S');
}
end;
(* end module getdatetime {for FPC} *)

(* begin module readdatetime *)
procedure readdatetime (var thefile: text; var adatetime: datetimearray);
(* read the date and time from the file.  It must have this format:
123456789 123456789
         1         
1980/06/09 18:49:11
 *)
(*
2001 Jan 2: thefile is not written to!  Force 4 digit years.
2000 Oct 11: upgraded so that the p2c compiler does not object
to writing out the adatetime; added checks for the date.
*)
var
   index: integer; (* to the udatetime *)
   (* the following is an unpacked date time array, to avoid reading
   into a packed array.  reading into a packed array is not transportable *)
   udatetime: array[1..datetimearraylength] of char;
begin
   for index:=1 to datetimearraylength do read(thefile,udatetime[index]);
   pack(udatetime, 1, adatetime);
   if (adatetime[3]='/') and (adatetime[12]=':') then begin
      writeln(output,'You have an old datetime (only 2 year digits): ');
      for index:=1 to datetimearraylength do write(output,adatetime[index]);
      writeln(output);
      writeln(output,'Convert your database to 4 digit years.');
      halt;
   end;
   (* check the adatetime format.  Note that further checks for the
      other positions in the array could be done to be sure that
      they are numbers.  But this should be pretty good. *)
   if (adatetime[ 5]<>'/') or
      (adatetime[ 8]<>'/') or
      (adatetime[14]<>':') or
      (adatetime[17]<>':')
   then begin
      writeln(output,'readdatetime: bad date time read:');
      for index:=1 to datetimearraylength do write(output,adatetime[index]);
      writeln(output);
      halt
   end;
end;
(* end module readdatetime *)

(* begin module writedatetime *)
procedure writedatetime(var thefile: text; adatetime: datetimearray);
(* expand the date and time out and print in the file *)
var
      index: integer; (* index of datetime *)
begin
      for index:=1 to datetimearraylength do write(thefile,adatetime[index])
end;
(* end module writedatetime version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module timeseed *)
(* Read the computer date and time.  Reverse the order of the digits and put
a decimal point in front.  This gives a fraction between zero and one that
varies quite quickly, and is always unique (if the computer has sufficient
accuracy).  It is to be used as a seed to a random number generator.  This
has the nice property that the seed changes every second and does not repeat
for thousands of years!  *)
procedure addtoseed(var seed, power: real; c: char);
(* add the digit represented by c to the seed at the power position *)
var
      n: integer; (* the character represented by c *)
begin (* addtoseed *)
      power := power/10;
{
writeln(output,'addtoseed, c = ',c);
writeln(output,'addtoseed, ord(c) = ',ord(c));
}
   n := ord(c) - ord('0');
   if (n < 0) or (n > 9) then begin
      writeln(output,'timeseed: error in datetime');
      writeln(output,'it contains "',c,'" which is not a number.');
      writeln(output,'The getdatetime routine must be fixed.');
      halt;
   end;
   seed := seed + power*n
end; (* addtoseed *)

procedure makeseed(adatetime: datetimearray; var seed: real);
(* convert adatetime to a real number in seed, reversed order
Here is the standard adatetime format:
123456789 123456789
         1         
1980/06/09 18:49:11
*)
var
   power: real; (* a digit of the seed such as 0.01 *)
begin
   seed := 0.0;
   power := 1.0;
   addtoseed(seed, power, adatetime[19]);
   addtoseed(seed, power, adatetime[18]);
   (* : *)
   addtoseed(seed, power, adatetime[16]);
   addtoseed(seed, power, adatetime[15]);
   (* : *)
   addtoseed(seed, power, adatetime[13]);
   addtoseed(seed, power, adatetime[12]);
   (*   *)
   addtoseed(seed, power, adatetime[10]);
   addtoseed(seed, power, adatetime[ 9]);
   (* / *)
   addtoseed(seed, power, adatetime[ 7]);
   addtoseed(seed, power, adatetime[ 6]);
   (* / *)
   addtoseed(seed, power, adatetime[ 4]);
   addtoseed(seed, power, adatetime[ 3]);
   addtoseed(seed, power, adatetime[ 2]);
   addtoseed(seed, power, adatetime[ 1]);
end;

procedure orderseed(adatetime: datetimearray; var seed: real);
(* convert adatetime to a real number in seed, normal order *)
var
   power: real; (* a digit of the seed such as 0.01 *)
begin
   seed := 0.0;
   power := 1.0;
   addtoseed(seed, power, adatetime[ 3]);
   addtoseed(seed, power, adatetime[ 4]);
   addtoseed(seed, power, adatetime[ 6]);
   addtoseed(seed, power, adatetime[ 7]);
   (* / *)
   addtoseed(seed, power, adatetime[ 9]);
   addtoseed(seed, power, adatetime[10]);
   (* / *)
   addtoseed(seed, power, adatetime[12]);
   addtoseed(seed, power, adatetime[13]);
   (*   *)
   addtoseed(seed, power, adatetime[15]);
   addtoseed(seed, power, adatetime[16]);
   (* : *)
   addtoseed(seed, power, adatetime[18]);
   addtoseed(seed, power, adatetime[19]);
end;

procedure timeseed(var seed: real);
(* read the computer date and time.  reverse the order of the digits
and put a decimal point in front.  this gives a fraction between
zero and one that varies quite quickly, and is always unique (if the
computer has sufficient accuracy).  it is to be used as a seed to
a random number generator. *)
var
   adatetime: datetimearray; (* a date and time *)
begin (* timeseed *)
   getdatetime(adatetime);
{
writeln(output,'timeseed: adatetime: ',adatetime);
}
   makeseed(adatetime, seed);
end; (* timeseed *)
(* end module timeseed version = 1.01; (@ of testtime.p 1997 Jan 11 *)

(*[[*)
(* begin module limitdate *)
procedure limitdate(a,b,c,d: char; limitdatetime: datetimearray);
(* test whether the current time is before the limit.
If it is later, halt the program *)
var
   adatetime: datetimearray; (* a date and time *)
   Dday: real; (* the critical day *)
   now: real;  (* this very moment *)
begin
   getdatetime(adatetime);
{
writeln(output,'adatetime:',adatetime);
writeln(output,'adatetime[1]: ', adatetime[1]);
writeln(output,'adatetime[2]: ', adatetime[2]);
writeln(output,'adatetime[3]: ', adatetime[3]);
writeln(output,'adatetime[4]: ', adatetime[4]);
writeln(output,'adatetime[5]: ', adatetime[5]);
}

   orderseed(adatetime, now);

   if
      (limitdatetime[1] <> ' ') or
      (limitdatetime[2] <> ' ') or
      (limitdatetime[3] <> ' ') or
      (limitdatetime[4] <> ' ') then halt;

   limitdatetime[1] := a;
   limitdatetime[2] := b;
   limitdatetime[3] := c;
   limitdatetime[4] := d;

   orderseed(limitdatetime, Dday);

{
   writeln(output,'now:  ',now:20:10);
   writeln(output,'Dday: ',Dday:20:10);
}
   if now > Dday then begin
       writeln(output,'This program expired on ',limitdatetime);
       writeln(output,'See: https://alum.mit.edu/www/toms/walker/contacts.html');
       halt
   end
end;
(* end module limitdate *)
(*]]*)

(* begin module demo.time *)
procedure demotime(var fout: text);
(* write the time to file fout *)
var   dateandtime: datetimearray; (* the date and time *)
      seed: real; (* a seed for a random number generater, made
         from the date and time written backwards *)
begin (* demotime *)
   getdatetime(dateandtime);
   writeln(fout);
   writeln(fout,'The date and time is: ');
   writedatetime(fout, dateandtime);
   writeln(fout,' <- This should be the current time and date.');
   write  (fout,'1980/06/09 18:49:11');
   writeln(fout,' <- Times and dates should look like this');
   writeln(fout,'year mo da ho mi se <- with parts in these positions');
   writeln(fout);

   timeseed(seed);
   writeln(fout,'A timeseed is ',seed:16:14);
   writeln(fout,'Timeseeds can be used to start random number generators.')
end; (* demotime *)
(* end module demo.time version = 7.40; {of delmod.p 2000 Feb 18} *)

(* begin module timefpc.themain *)
procedure themain;
(* the main procedure of the program *)
begin
   writeln(output,'timefpc ',version:4:2);

   demotime(output);
end;
(* end module timefpc.themain *)

begin
   themain;
1: end.
