program todobfcite(input, todobfcitep, todobfcitelist, output);
(* todobfcite: find cite inside todobf of a LaTeX document

  Thomas D. Schneider, Ph.D.
  toms@alum.mit.edu
  toms@alum.mit.edu (permanent)
  https://alum.mit.edu/www/toms (permanent)

 *)

label 1; (* end of program *)

const
(* begin module version *)
   version = 1.06; (* of todobfcite.p 2014 Jan 29
2014 Jan 29, 1.06: 
2014 Jan 29, 1.05: working on handle multiple cites inside todobf
2014 Jan 08, 1.04: handle multiple cites inside todobf (goal)
2014 Jan 08, 1.03: clean up.
2014 Jan 08, 1.02: building ... FULLY FUNCTIONAL!!!
2014 Jan 08, 1.01: detects todobf{} region!
2014 Jan 07, 1.00: origin (copies input to output)
*)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.todobfcite *)
(*
name
   todobfcite: find cite inside todobf of a LaTeX document

synopsis
   todobfcite(input: in, todobfcitep: in, todobfcitelist: out, output: out)

files

   input:  a LaTeX document containing \todobf{} and \cite{} inside
     the todobfs.

   todobfcitelist:  not used.

   todobfcitep:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

      not used. (yet?)

   output: messages to the user

description

   The 'reading' script looks for todobf commands that surround cite
   commands:

   \todobf{ ... \cite{...} ...}

   and identifies the keys for those papers.  Then it copies the PDFs
   of those papers into the dropbox so that they can be read on my
   iPad.

   Although the reading script does this, it is slow.  This Pascal
   program ought to be much faster.

examples

documentation

see also

   reading

author

   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.todobfcite *)

(* begin module string.const *)
      maxstring = 2000; (* the maximum string *)
(* end module string.const version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module filler.const *)
      fillermax = 50; (* the size of the filler array for a string *)
(* end module filler.const version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

type

(* begin module string.type *)
   stringptr = ^string; (* pointer to a string *)
   string = record (* a string of characters *)
      letters: array[1..maxstring] of char; (* the letters in the string *)
      length: integer; (* the number of characters in the string *)
      current: integer; (* the letter we are working on *)
      previous: stringptr; (* the previous string in a series *)
      next: stringptr; (* the next string in a series *)
   end;
(* end module string.type version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module filler.type *)
      (* the following is an array used to fill a string.
      it is convenient to have it much shorter than the maxstring, so that
      it is easy to fill the string using procedure fillstring.
      the user must declare the value of constant fillermax. *)
      filler = packed array[1..fillermax] of char;
(* end module filler.type version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module trigger.type *)
      trigger = record (* an object to be searched for *)
         seek: string; (* the characters looked for *)
         state: integer; (* how close to triggering we are *)
         skip: boolean; (* trigger not found- skip the line *)
         found: boolean (* the trigger was found *)
      end;
(* end module trigger.type version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

var
   todobfcitep, (* file used by this program *)
   todobfcitelist: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module package.trigger *)
(* ************************************************************************ *)

(* begin module clearstring *)
(* These modules clear strings in various ways *)
(* ---- *)
procedure emptystring(var ribbon: string);
(* empty the contents of the string but do NOT
remove the pointer.  This is useful for clearing
one string within a linked list of them. *)
var   index: integer; (* to the ribbon *)
begin (* clearstring *)
   with ribbon do begin
      for index := 1 to maxstring do letters[index] := ' ';
      length := 0;
      current := 0;
   end
end; (* emptystring *)
(* ---- *)
procedure clearstring(var ribbon: string);
(* empty the string and remove the pointer.
As of 2010 Dec 04 there is a previous pointer that
allows proper removal of the linked list using disposestring.
This clearstring routine is for the initial clearing
of the variables. *)
begin (* clearstring *)
   with ribbon do begin
      emptystring(ribbon);
      previous := nil;
      next := nil;
   end
end; (* clearstring *)
(* ---- *)
procedure disposestring(var ribbon: stringptr);
(* remove the linked list from the string in revere order *)
var
   t: stringptr; (* a link in the string *)
begin
   t := ribbon;
   (* find the end *)
   while t^.next <> nil do t := t^.next;
   (* move backwards and dispose links *)
   while t^.previous <> nil do begin
      t := t^.previous;
      dispose(t^.next);
   end;
   dispose(t);
end; (* disposestring *)
(* ---- *)
procedure initializestring(var ribbon: string);
(* start the string with a nil pointer.   This routine
should be called before doing linked list work.  This allows
the standard string routines to clear the string without
killing the pointer.  This is now deprecated, do not use it
since clearstring still clears the next pointer. *)
begin (* initializestring *)
   writeln(output,'remove initializestring routine!');
   writeln(output,'replace it with clearstring routine!');
   halt; (* to force deprecation *)
   clearstring(ribbon);
   ribbon.next := nil;
end; (* initializestring *)
(* end module clearstring version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module writestring *)
procedure writestring(var tofile: text; var s: string);
(* write the string s to file tofile, no writeln *)
var   i: integer; (* index to s *)
begin (* writestring *)
      with s do for i := 1 to length do write(tofile, letters[i])
end; (* writestring *)
(* end module writestring version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module filler.fillstring *)
procedure fillstring(var s: string; a: filler);
(* this procedure makes it reasonably easy to fill the string s with
characters.  one calls the procedure as: *)
(*                           1         2         3         4         5 *)
(*                  12345678901234567890123456789012345678901234567890 *)
(*   fillstring(s, 'this-is-the-string                                ');
the two comments make it easy to line the characters up. also, for this
example, it was assumed that the length of filler as defined by the
constant fillermax was 50. *)
var
      length: integer; (* of the string without trailing blanks *)
      index: integer; (* of s *)
begin (* fillstring *)
      clearstring(s);
      length := fillermax;
      while (length > 1) and (a[length] = ' ') do length := pred(length);
      if (length = 1) and (a[length] = ' ') then begin
         writeln(output, 'fillstring: the string is empty');
         halt
      end;

      for index := 1 to length do s.letters[index] := a[index];
      s.length := length;
      s.current := 1
end; (* fillstring *)
(* end module filler.fillstring version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module filler.filltrigger *)
procedure filltrigger(var t: trigger;
                      a: filler);
(* fill the trigger t *)
begin (* filltrigger *)
         fillstring(t.seek,a)
end; (* fillstring *)
(* end module filler.filltrigger version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module trigger.proc *)
(* this module allows one to scan a series of characters, as from
an array or a file, and to "trigger" or detect a simple string
in the series.  the advantage of the trigger is that several triggers
can "observe" a stream of characters at once, each looking for a
different thing.
some other modules required: interact.const, interact.type *)

procedure resettrigger(var t: trigger);
(* reset the trigger to ground state *)
begin (* resettrigger *)
      with t do begin
         state := 0;
         skip := false;
         found := false
      end
end; (* resettrigger *)

procedure testfortrigger(ch: char; var t: trigger);
(* look at the character ch.
   if it is part of the trigger (at the current trigger state),
       then the trigger state goes higher.
   if it is not part of the trigger then the trigger state is reset,
      skip is true and one should skip onward to find the trigger.
   if the trigger is found, found is true.
1996 Sep 12: Bug found!  In the case of a trigger "ab", the program
used to miss it for situations like "aab".  This was because at the
first a it would step up.  Then it would see the second a and recognize
that was not part of ab.  It would fail to realize that it could be
the start of a new one.  The code now accounts for that possibility. *)
begin (* testfortrigger *)
      with t do begin
         state := succ(state);
{
writestring(list,seek);
writeln(list,'testfortrigger seek.letters[',state:1,']:',
seek.letters[state],' ch:',ch);
}
         if seek.letters[state] = ch
         then begin
            skip := false;
            if state = seek.length then found := true
                                   else found := false
         end
         else begin
            (* it failed.  But wait!  It could be the beginning
            of a NEW trigger string! *)
            if seek.letters[1] = ch then begin
               state := 1;
               skip := false;
               found := false
            end
            else begin (* reset trigger *)
               state := 0;
               skip := true;
               found := false
            end
         end
      end
end; (* testfortrigger *)
(* end module trigger.proc version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module skipblanks *)

(* 2003 July 31: tab is considered a blank character *)

function isblank(c: char): boolean;
(* is the character c blank or tab? *)
const
   tab = 9; (* tab character *)
begin
   isblank := (c = ' ') or (ord(c) = tab)
end;

procedure skipblanks(var thefile: text);
(* skip over blanks until a non-blank, or end of line, is found *)
begin
   while isblank(thefile^) and not eoln(thefile) do get(thefile);
end;

procedure skipnonblanks(var thefile: text);
(* skip over nonblanks until a blank, or end of line, is found *)
begin
   while (not isblank(thefile^)) and not eoln(thefile) do get(thefile);
end;

procedure skipcolumn(var thefile: text);
(* skip over a data column *)
begin
   skipblanks(thefile); skipnonblanks(thefile)
end;
(* end module skipblanks version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* ************************************************************************ *)
(* end module package.trigger version = 5.52; (@ of prgmod.p 2013 Jun 10 *)

(* begin module todobfcite.themain *)
procedure themain(var input, todobfcitep, todobfcitelist: text);
(* the main procedure of the program *)
var
{
   parameterversion: real; (* parameter version number *)
}
   c: char; (* a character from the input *)
   todobf: trigger; (* a string to search for *)
   cite: trigger; (* a character from the input *)
   intodobf: boolean; (* true when inside a todobf *)
   incite: boolean; (* true when inside a cite *)
   ptodobf: integer; (* count of curly parenthesis '{' and '}' under todobf *)
   pcite: integer; (* count of curly parenthesis '{' and '}' under cite*)
   doverbose: boolean; (* show the output *)
   justbroke: boolean; (* we just broke the line *)
begin
{
   writeln(output,'todobfcite ',version:4:2);
   reset(todobfcitep);
   readln(todobfcitep, parameterversion);
   if round(100*parameterversion) < round(100*updateversion) then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;
}

   doverbose := false;
   doverbose := true;

   (* set up triggers *)
(*                            1         2         3         4         5 *)
(*                   12345678901234567890123456789012345678901234567890 *)
filltrigger(todobf, '\todobf                                           ');
filltrigger(cite  , '\cite                                             ');
   resettrigger(todobf);
   resettrigger(cite);
   intodobf := false;
   incite := false;
   ptodobf := 0;
   pcite := 0;

   justbroke := false;
   rewrite(todobfcitelist);
   while not eof(input) do begin
      if eoln(input) then begin
         readln(input);
         writeln(output);
      end
      else begin (* read characters of a line *)
         read(input,c);
         write(output,c);

         testfortrigger(c, todobf);
         if todobf.found then begin
            intodobf := true;
            ptodobf := 0;
         end
         else begin (* only test if todobf.found is not still true *)
            if intodobf and (c = '{') then ptodobf := ptodobf + 1;
         end;
         if intodobf and (c = '}') then ptodobf := ptodobf - 1;
         if (not todobf.found) and intodobf and (ptodobf = 0) then begin
            (* this is the end of the todobf *)
            intodobf := false;
            if doverbose then write(output,'[*** END OF todobf ***]');
         end;

         if intodobf then begin
            testfortrigger(c, cite);
            if cite.found then begin
               incite := true;
               pcite := 0;
            end
            else begin (* only test if cite.found is not still true *)
               if incite and (c = '{') then pcite := pcite + 1;
            end;

            if incite and (c = '}') then begin
               (* after the citation is done! *)
               pcite := pcite - 1;
               (* This would mark the end of every citation set:
               if not justbroke then writeln(todobfcitelist, 'SNAPPPPPPPPP');
               *)
            end;

            if (not cite.found) and incite and (pcite = 0) then begin
               (* this is the end of the cite *)
               incite := false;
               writeln(todobfcitelist);
               justbroke := true;
               if doverbose then write(output,'[*** END OF cite ***]');
            end;
         end;

         if intodobf and incite and (pcite = 1) and (c <> '{') then begin
            if (c <> ',') and (c <> ' ') then begin
               (* This is what writes the list of citations *)
               write(todobfcitelist, c);
               justbroke := false;
            end;

(*
Example of breaks:
\cite{Smith.Loiselle2005, Lichtwark.Barclay2010, Barclay.Curtin2010}.
*)
            (* are we at a break? *)
            if (c = ',') or (c = ' ') then begin
               (* This breaks the references into individual lines *)
               if not justbroke then writeln(todobfcitelist);
               justbroke := true;
            end;
         end;

      end
   end;

end;
(* end module todobfcite.themain *)

begin
   themain(input, todobfcitep, todobfcitelist);
1: end.
