program tram(fin, tramp, fout, output);
(* tram: transfac matrix analysis

  Tom Schneider
  toms@alum.mit.edu
  https://alum.mit.edu/www/toms/


 *)

label 1; (* end of program *)

const
(* begin module version *)
version = 1.01; (* of tram.p 1997 April 22
origin 1997 April 17 *)
updateversion = 1.00; (* defines lowest acceptable current parameter file *)
(* end module version *)

(* begin module describe.tram *)
(*
name
   tram: transfac matrix analysis

synopsis
   tram(fin: in, tramp: in, fout: out, output: out)

files
   fin:  the transfrac matrix.dat file

   fout:  crude information analysis of each matrix.

   tramp:  parameters to control the program.  The file must contain the
      following parameters, one per line:

      parameterversion: The version number of the program.  This allows the
         user to be warned if an old parameter file is used.

   output: messages to the user

description

   Crude analysis (no small sample correction) of transfac database.

   Note:  In Dfd are normalized numbers that sum to 4 - idiots.
   This means that the tables are real numbers, not integers.

examples

documentation

see also
   trex.p

author
   Thomas Dana Schneider

bugs

technical notes

*)
(* end module describe.tram *)

var
   fin, (* file used by this program *)
   tramp, (* file used by this program *)
   fout: text; (* file used by this program *)

(* begin module halt *)
procedure halt;
(* stop the program.  the procedure performs a goto to the end of the
   program.  you must have a label:
      label 1;
   declared, and also the end of the program must have this label:
      1: end.
   examples are in the module libraries.
   this is the only goto in the delila system. *)
begin
      writeln(output,' program halt.');
      goto 1
end;
(* end module halt version = 'delmod 6.16 84 mar 12 tds/gds'; *)

(* begin module copynoreturn *)
procedure copynoreturn(var fin, fout: text);
(* copy a line from file fin to file fout
but don't put a carriage return *)
begin (* copynoreturn *)
      while not eoln(fin) do begin
         fout^ := fin^;
         put(fout);
         get(fin)
      end;
{
      readln(fin);
}
(*      writeln(fout); *)
end; (* copynoreturn *)
(* end module copynoreturn *)

function plogp(f: real): real;
(* compute p log2 p for f *)
begin
   if f <= 0.0
   then plogp := 0.0
   else plogp := -f * ln(f)/ln(2)
end;

(* begin module tram.themain *)
procedure themain(var fin, tramp, fout: text);
(* the main procedure of the program. *)
const
   maxcount = 2000;
   wid = 4;
   dec = 1;
var
   count: integer; (* count of loops done *)
   done: boolean; (* done with looping *)
   H: real; (* H function of na, nc, ng, nt *)
   id: array[1..2] of char; (* id for database lines *)
   i: integer; (* index for printing information curve *)
   na,nc,ng,nt: real; (* the numbers of 4 bases *)
   n: real; (* sum of na, nc, ng, nt *)
   parameterversion: real; (* parameter version number *)
   R: real; (* 2- H *)
begin
   writeln(output,'tram ',version:4:2);
   reset(tramp);
   readln(tramp, parameterversion);
   if parameterversion < updateversion then begin
      writeln(output, 'You have an old parameter file!');
      halt
   end;

   reset(fin);
   rewrite(fout);

   count := 0;
   done := false;
   while not done do begin
      if eof(fin) or (count >= maxcount)
      then done := true
      else begin
         read(fin, id);
{
writeln(fout,'id = ',id);
}
         if id = 'NA' then begin
            write(fout,'count = ',count:1);
            copynoreturn(fin,fout);
            writeln(fout);
         end;
         readln(fin);

         if id = 'P0' then begin
            count := succ(count);
{
writeln(fout,'count = ',count:1);
}
            while id <> 'XX' do begin
               read(fin, id);
{
writeln(fout,'id = ',id,' inner');
}
               if id <> 'XX' then begin
                  read(fin, na, nc, ng, nt);
                  write(fout, ' ',na:wid:dec, ' ',nc:wid:dec,
                              ' ',ng:wid:dec, ' ',nt:wid:dec);

                  copynoreturn(fin,fout);
                  write(fout,' | ');

                  n := na+nc+ng+nt;
                  H :=   plogp(na/n)
                       + plogp(nc/n)
                       + plogp(ng/n)
                       + plogp(nt/n);
                  R := 2 - H;  (* no sample size correction *)

{
                  write(fout,' H ',H:5:2);
}
                  write(fout,' R ',R:5:2);
                  write(fout,' | ');
                  for i := 0 to round(20*R) do write(fout,'*');
                  write(fout,'|');
 
                  writeln(fout);
               end;
               readln(fin);
            end
         end
      end
   end;

end;
(* end module tram.themain *)

begin
   themain(fin, tramp, fout);
1: end.




























