program wwfmt(input, output);
(* wwfmt: word wrap like fmt
   Tom Schneider *)

const
(* begin module version *)
version = 1.01; (* of wwfmt.p 2000 Nov 12
origin 2000 Nov 12 from ww *)
(* end module version *)

(* begin module describe.ww *)
(*
name

   wwfmt: word wrap like fmt

synopsis

   ww(input: in, output: out)

files

   input:  text to be wrapped

   output: wrapped text

description

This Pascal program takes ASCII text and filters it.  Lines longer than the
constant maxline are altered by replacing the first space after position
maxline with a carriage return.  This has the effect of wrapping the lines
between 'words'.

Unlike the ww program, wrapping continues for a paragraph.  Paragraphs
are separated by blank lines with no characters on them.

The program simulates the unix fmt but unfortunately can't
read the command line.

examples

documentation

see also
   ww.p
   wl.p

author

   Thomas Dana Schneider

bugs

   The constant maxline is fixed at compile time, of course.
   This probably can be done using a gpc version eventually.

technical notes

*)
(* end module describe.ww *)

    maxline = 65; (* maximum width of output lines *)

var c: char; (* a character in the input *)
    n: integer; (* number of characters written to output so far *)
    waitforspace: boolean; (* true if waiting for the next space to appear *)
    paragraph: boolean; (* true if we just did a paragraph *)
begin

   n := 0;
   paragraph := false;
   while not eof(input) do begin
{
write(output,'>');
}
      if eoln(input) then begin
         readln(input);
         if not eof(input) then begin
            if eoln(input) then begin (* paragraph break *)
               writeln(output);
               if not paragraph then writeln(output);
               paragraph := true;
               n := 0; (* no characters written out so far on this line *)
            end
            else begin
               if n > 0 then begin
                  write(output,' ');  (* Separate items on the line *)
                  n := n + 1;
                  paragraph := false;
{
write(output,'/',n:1,'/');  (* Separate items on the line *)
}
               end;
            end;
         end;
      end
      else begin
         waitforspace := false;
         while not eoln(input) do begin
            if n > maxline then waitforspace := true;
            read(input, c);

            if waitforspace and (c = ' ')
            then begin 
               writeln(output,' '); (* replace space with carriage return *)
               n := 0; (* no characters written out so far on this line *)
{
write(output,'#');
}
               waitforspace := false;
            end
            else begin
              write(output, c);
              n := n + 1;
            end

         end;
      end
   end;

   if n <> 0 then begin
      writeln(output);
   end;

end.

